'use strict';

const _Ok = (arg) => ({
    ap: (r) => (typeof arg === "function" ? r.map((x) => arg(x)) : Err()),
    map: (cb) => _Ok(cb(arg)),
    mapErr: () => _Ok(arg),
    chain: (cb) => cb(arg),
    chainErr: () => _Ok(arg),
    swap: () => _Err(arg),
    bimap: (ok, _) => _Ok(ok(arg)),
    cata: (obj) => obj.Ok(arg),
    toMaybe: () => nullable(arg),
    inspect: () => `Ok(${arg})`,
    isErr: () => false,
    isOk: () => true,
    value: () => arg,
});
const _Err = (arg) => ({
    ap: () => _Err(arg),
    map: () => _Err(arg),
    mapErr: (cb) => _Err(cb(arg)),
    chain: () => _Err(arg),
    chainErr: (cb) => cb(arg),
    swap: () => _Ok(arg),
    bimap: (_, err) => _Err(err(arg)),
    cata: (obj) => obj.Err(arg),
    toMaybe: () => Nothing,
    inspect: () => `Err(${arg})`,
    isErr: () => true,
    isOk: () => false,
    value: () => arg,
});
function Ok(arg) {
    return _Ok(arg);
}
function Err(arg) {
    return _Err(arg);
}

const Just = (arg) => ({
    ap: (m) => (typeof arg === "function" ? m.map((v) => arg(v)) : Nothing),
    map: (cb) => Just(cb(arg)),
    chain: (cb) => cb(arg),
    alt: () => Just(arg),
    cata: (obj) => obj.Just(arg),
    toResult: () => Ok(arg),
    inspect: () => `Just(${arg})`,
    isNothing: () => false,
    isJust: () => true,
    value: () => arg,
});
const Nothing = {
    ap: () => Nothing,
    map: () => Nothing,
    chain: () => Nothing,
    alt: (a) => Just(a),
    cata: (obj) => obj.Nothing(),
    toResult: () => Err(),
    inspect: () => `Nothing`,
    isNothing: () => true,
    isJust: () => false,
    value: () => undefined,
};
const nullable = (arg) => (arg === null || arg === undefined ? Nothing : Just(arg));

const encase = (func) => {
    try {
        return Just(func());
    }
    catch {
        return Nothing;
    }
};
const encaseRes = (func) => {
    try {
        return Ok(func());
    }
    catch (e) {
        return Err(e);
    }
};

const get = (selector) => (data) => selector.reduce((acc, s) => acc.chain((d) => nullable(d[s])), nullable(data));

const head = (arr) => (arr.length ? Just(arr[0]) : Nothing);

// justs: takes an array of maybes and filters out any non-just types
const justs = (arr = []) => arr.filter((a) => a.isJust());

const last = (arr) => (arr.length ? Just(arr[arr.length - 1]) : Nothing);

// oks: takes an array of Results and filters out any non-ok types
const oks = (arr = []) => arr.filter((a) => a.isOk());

const parseDate = (date) => nullable(date)
    .map((d) => new Date(d))
    .chain((d) => (isNaN(d.valueOf()) ? Nothing : Just(d)));

const tail = (arr) => (arr.length ? Just(arr.slice(1)) : Nothing);

const tryFind = (exp) => (data) => nullable(data.find(exp));

const collectResult = (results) => {
    const successes = [];
    const failures = [];
    for (const result of results) {
        result.cata({
            Ok: (x) => successes.push(x),
            Err: (x) => failures.push(x),
        });
    }
    if (failures.length > 0) {
        return Err(failures);
    }
    return Ok(successes);
};
const collectMaybe = (maybes) => {
    const values = [];
    for (const maybe of maybes) {
        if (maybe.isNothing()) {
            return Nothing;
        }
        values.push(maybe.value());
    }
    return Just(values);
};

exports.Err = Err;
exports.Just = Just;
exports.Nothing = Nothing;
exports.Ok = Ok;
exports.collectMaybe = collectMaybe;
exports.collectResult = collectResult;
exports.encase = encase;
exports.encaseRes = encaseRes;
exports.get = get;
exports.head = head;
exports.justs = justs;
exports.last = last;
exports.nullable = nullable;
exports.oks = oks;
exports.parseDate = parseDate;
exports.tail = tail;
exports.tryFind = tryFind;
